/*
 * Copyright (c) 2011 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file global_equalization.c
 * @brief global histogram equalization
 *
 * @author Jose-Luis Lisani <joseluis.lisani@uib.es>
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>

/* Auxiliary data structure for Global Equalization */
struct data_float {
    float value;
    int pixel;
};

/**
 * @brief sort float data in increasing order
 *
 */
int order_increasing_data_float(const void *pVoid1, const void *pVoid2) {
    struct data_float *node1, *node2;

    node1=(struct data_float *) pVoid1;
    node2=(struct data_float *) pVoid2;

    if (node1->value < node2->value) return -1;
    if (node1->value > node2->value) return 1;
    return 0;
}

/**
 * @brief sort input data
 *
 * This function sorts the input data in increasing order
 *
 * @param data input array
 * @param size array size
 * @return F, sorted values
 */
void
sort_data_pixel(float *data, size_t size, struct data_float *F) {
    int n;

    //copy original data and position
    for (n=0; n < (int) size; n++) {
        F[n].value=data[n];
        F[n].pixel=n;
    }

    //order in increasing order
    qsort(F, size, sizeof(struct data_float), order_increasing_data_float);

}


/**
 * @brief Global histogram equalization between 0.0 and 255.0
 *
 *
 * @param data_out the array with the transformed initial array
 * @param data initial array
 * @param dim size of the array
 *
 */
int
global_equalization(float *data_out, float *data, size_t dim) {
    struct data_float *F;
    int             n, nnext, m, pos;
    double          constantF;

    if (NULL == (F = (struct data_float *) \
                     malloc(dim * sizeof(struct data_float)))) {
        fprintf(stderr, "allocation error\n");
        free(data);
        return EXIT_FAILURE;
    }

    /* sort input data in increasing order,
       keep pixel location */
    sort_data_pixel(data, dim, F);

    constantF=255.0/(double) dim;
    for (n=0; n < (int)  dim; n=nnext) {
        /* pos = maximum index of sorted data such that */
        /*       value_sorted[n] <= value_sorted[pos]   */
        pos=n;
        m=n+1;
        while ((m < (int) dim) && (F[m].value == F[n].value)) {
            pos=m;
            m=m+1;
        }
        /* assign new value to corresponding pixels */
        m=n;
        while ((m < (int) dim) && (F[m].value == F[n].value)) {
            data_out[F[m].pixel]=(float) ((double) (pos+1) * constantF);
            m=m+1;
        }
        nnext=m;
    }

    free(F);
    return EXIT_SUCCESS;

}

