/*
 * Copyright (c) 2011 Catalina Sbert <catalina.sbert@uib.es>
 * All rights reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @mainpage Piecewise Affine Equalization
 *
 * README.txt:
 * @verbinclude README.txt
 */

/**
 * @file piecewise_equalization.c
 * @brief piecewise afine transformation
 *
 *
 *
 * @author Catalina Sbert <catalina.sbert@uib.es>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "io_png.h"
#include "piecewise_equalization_lib.h"

int
main(int argc, char *const * argv) {

    float           smax;       /* maximum slope of the affine transformation */
    float           smin;       /* minimum slope of the affine transformation */
    int             N;          /* number of intervals  of the partition */
    size_t          nx, ny, nc, dim;
    float          *data, *data_out=NULL, *data_outc=NULL, *ptr_red, *ptr_green,
    *ptr_blue;
    float          *data_gray, *data_gray1=NULL, *ptr_gray, *ptr_end;
    int             channel, n;
    FILE           *points;
    char           *namedataout;




    /* wrong number of parameters : simple help info */
    if (7 > argc) {
        fprintf(stderr, "usage : %s smax smin N in.png  out_g.png out_c.png \
 [namedataout]\n", argv[0]);
        fprintf(stderr, "        smax slope [ >= 1] smin slope \
 [0, 1] N subintervals [-1 (global equalization) or N >= 2]\n");
        return EXIT_FAILURE;
    }
    if (8 != argc) {
        /* default output file name */
        namedataout = "points.dat";
    } else {
        namedataout = argv[7];
    }
    smax = atof(argv[1]);
    if (smax < 1) {
        fprintf(stderr, "the maximum slope cannot be smaller than 1\n");
        return EXIT_FAILURE;
    }

    smin = atof(argv[2]);
    if ((smin < 0) || (smin > 1)) {
        fprintf(stderr, "the  minimum slope must be between 0 and 1\n");
        return EXIT_FAILURE;
    }

    if (smin > smax) {
        fprintf(stderr, "the minimum slope cannot be greater than the \
                maximum slope\n");
        return EXIT_FAILURE;
    }

    /* read the PNG image into data */
    if (NULL == (data = io_png_read_f32(argv[4], &nx, &ny, &nc))) {
        fprintf(stderr, "the image could not be properly read\n");
        return EXIT_FAILURE;
    }
    dim = nx * ny;

    N = atoi(argv[3]);
    if ((N != -1) && ((N < 2) || (N > (int) dim))) {
        fprintf(stderr, "the  number of subintervals must be between 2 and the \
image size, or -1 for global equalization\n");
        free(data);
        return EXIT_FAILURE;
    }
    if (N != -1)
        N = N - 1;              /* number of intermediate points */


    points = fopen(namedataout, "w");




    if (3 <= nc)
        channel = 3;
    else
        channel = 1;

    /* allocate data_out  */
    if (NULL == (data_out = (float *) malloc(nc * dim * sizeof(float))) ||
            NULL == (data_outc = (float *) malloc(nc * dim * sizeof(float)))) {
        fprintf(stderr, "allocation error\n");
        free(data);
        if (data_out) free(data_out);
        if (data_outc) free(data_outc);
        return EXIT_FAILURE;
    }
    /* allocate data_gray */
    if ((NULL == (data_gray = (float *) malloc(dim * sizeof(float)))) ||
            (NULL == (data_gray1 = (float *) malloc(dim * sizeof(float))))) {
        fprintf(stderr, "allocation error\n");
        free(data);
        if (data_gray) free(data_gray);
        if (data_gray1) free(data_gray1);
        return EXIT_FAILURE;
    }



    if (channel == 3) {
        ptr_gray = data_gray;
        ptr_end = ptr_gray + dim;
        ptr_red = data;
        ptr_green = data + dim;
        ptr_blue = data + 2 * dim;
        while (ptr_gray < ptr_end) {
            *ptr_gray = (*ptr_red + *ptr_green + *ptr_blue) / 3.;
            ptr_gray++;
            ptr_red++;
            ptr_green++;
            ptr_blue++;
        }
    } else
        memcpy(data_gray, data, dim * sizeof(float));


    if (N != -1)
        piecewise_transformation(data_gray1, data_gray, dim, N, smax, smin,
                                 points);
    else
        global_equalization(data_gray1, data_gray, dim);

    if (channel == 3) {

        color_image(data_out, data, data_gray, data_gray1, dim);
        io_png_write_f32(argv[5], data_out, nx, ny, nc);


        for (n = 0; n < channel; n++) {
            fprintf(points, "channel %d\n", n);
            if (N != -1)
                piecewise_transformation(data_outc + n * dim, data + n * dim, dim, N, smax,
                                         smin, points);
            else
                global_equalization(data_outc + n * dim, data + n * dim, dim);

        }
        io_png_write_f32(argv[6], data_outc, nx, ny, nc);
    } else {
        io_png_write_f32(argv[5], data_gray1, nx, ny, nc);
        io_png_write_f32(argv[6], data_gray1, nx, ny, nc);
    }
        
    fclose(points);
    free(data_out);
    free(data_gray);
    free(data_gray1);
    free(data_outc);
    free(data);

    return EXIT_SUCCESS;

}
