/*
 * Copyright (c) 2011 Catalina Sbert <catalina.sbert@uib.es>
 * All rights reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file piecewise_equalization_lib.c
 * @brief routines of piecewise equalization
 *
 * @author Catalina Sbert <catalina.sbert@uib.es>
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <limits.h>

/**
 * @brief get the min/max of a float array
 *
 * @param data input array
 * @param size array size
 * @param ptr_min, ptr_max pointers to the returned values, ignored if NULL
 */
static void
minmax(const float *data, size_t size,
       float *ptr_min, float *ptr_max) {
    const float    *ptr_data, *ptr_end;
    float           min, max;

    /* sanity check */
    if (NULL == data) {
        fprintf(stderr, "a pointer is NULL and should not be so\n");
        abort();
    }

    /* compute min and max */
    ptr_data = data;
    ptr_end = ptr_data + size;
    min = *ptr_data;
    max = *ptr_data;
    ptr_data++;
    while (ptr_data < ptr_end) {
        if (*ptr_data < min)
            min = *ptr_data;
        if (*ptr_data > max)
            max = *ptr_data;
        ptr_data++;
    }

    /* save min and max to the returned pointers if available */
    if (NULL != ptr_min)
        *ptr_min = min;
    if (NULL != ptr_max)
        *ptr_max = max;
    return;
}

/**
 * @brief sort float data in increasing order
 *
 */
int order_increasing_float(const void *pVoid1, const void *pVoid2) {
    float *v1, *v2;

    v1=(float *) pVoid1;
    v2=(float *) pVoid2;

    if (*v1 < *v2) return -1;
    if (*v1 > *v2) return 1;
    return 0;
}


/**
 * @brief sort input data
 *
 * This function sorts the input data in increasing order
 *
 * @param data input array
 * @param size array size
 * @return F, sorted values
 */
void
sort_data(float *data, size_t size, float *F) {
    //copy original data
    memcpy(F, data, size*sizeof(float));

    //order in increasing order
    qsort(F, size, sizeof(float), order_increasing_float);

}

/**
 * @brief inverse cumulative distribution function
 *
 * Given a value Fu, computes the inverse of the cumulative histogram
 * by searching the appropriate value in the array of sorted data
 *
 * @param Fu  value (> 0)
 * @param F  sorted data
 * @return x the inverse value of Fu
 */

float
inverse_cumulative_function(float Fu, float *F) {
    float x;
    int pos;

    if (Fu < 0) {
        fprintf(stderr, "incorrect input value\n");
        abort();
    }

    pos=(int) ceil((double) Fu);
    pos=pos-1; /*array indexes start at 0*/

    x=F[pos];

    return x;
}

/**
* @brief Affine Transformation
*
* This function is the affine transformation of the interval [x0,x1] into
* [y0,y1]
*
* @param data input array
* @param dim dimension of the array
* @param x0, x1 initial interval
* @param y0, y1 transformed interval

@return data_out output array
*/


void
affine_transformation(float *data_out, float *data, size_t dim,
                      float x0, float x1, float y0, float y1) {
    float          *ptr_data, *ptr_data_out, *ptr_end;
    float           slope;

    slope = (y1 - y0) / (x1 - x0);

    ptr_data = data;
    ptr_end = data + dim;
    ptr_data_out = data_out;
    while (ptr_data < ptr_end) {
        if (x0 <= *ptr_data && *ptr_data <= x1) {
            *ptr_data_out = y0 + slope * (*ptr_data - x0);
            if (*ptr_data_out > 255.)
                 *ptr_data_out = 255.;
            if (*ptr_data_out < 0.)
                *ptr_data_out = 0.;
        }
        ptr_data++;
        ptr_data_out++;
    }


}
/**
* @brief computes the R G B components of the output image from its gray level
*
* Given a color image C=(R, G, B), given its gray level
*
* @f$ gray= (R+G+ B)/3 \f$
*
* Given a modified gray image gray1
*
* This function compute an output color image C1=(R1,G1,B1) which each channel
* is proportional to the input channel and whose gray level is gray1,
*
* @f$ R1=\frac{gray1}{gray} R
*     G1=\frac{gray1}{gray} G
*     B1= \frac{gray1}{gray} B \f$
*
* @param data_out Output color image
* @param data input color image
* @param gray gray level of the input color image
* @param gray1 modified gray image
* @param dim size of the image
*
* @return data_out
*/

void
color_image(float *data_out, float *data, float *gray, float *gray1,
            size_t dim) {

    float          *ptr_red, *ptr_green, *ptr_blue;
    float          *ptr_end, *ptr_gray, *ptr_gray1;
    float          *ptr_in_red, *ptr_in_green, *ptr_in_blue;
    float           A, B;

    /* sanity check */
    if (NULL == data_out || NULL == data || NULL == gray || NULL == gray1) {
        fprintf(stderr, "a pointer is NULL and should not be so\n");
        abort();
    }



    ptr_red = data_out;
    ptr_green = data_out + dim;
    ptr_blue = data_out + 2 * dim;
    ptr_gray = gray;
    ptr_gray1 = gray1;
    ptr_end = ptr_gray + dim;
    ptr_in_red = data;
    ptr_in_green = data + dim;
    ptr_in_blue = data + 2 * dim;
    while (ptr_gray < ptr_end) {
        if (*ptr_gray <= 1.)
            *ptr_gray = 1.;
        A = *ptr_gray1 / *ptr_gray;
        *ptr_red = A * (*ptr_in_red);
        *ptr_green = A * (*ptr_in_green);
        *ptr_blue = A * (*ptr_in_blue);
        if (*ptr_red > 255. || *ptr_green > 255. || *ptr_blue > 255.) {
            B = *ptr_in_red;
            if (*ptr_in_green > B)
                B = *ptr_in_green;
            if (*ptr_in_blue > B)
                B = *ptr_in_blue;
            A = 255. / B;
            *ptr_red = A * (*ptr_in_red);
            *ptr_green = A * (*ptr_in_green);
            *ptr_blue = A * (*ptr_in_blue);
        }
        ptr_gray++;
        ptr_gray1++;
        ptr_red++;
        ptr_green++;
        ptr_blue++;
        ptr_in_red++;
        ptr_in_green++;
        ptr_in_blue++;
    }
}


/**
* @brief Main block of the algorithm
*
* Given a regular partition of [0,255] with N {y_k} control points
* (N=number of intervals of the partition - 1), compute a new partition
* as the inverse of the cumulative histogram of the initial data  {x_k}.
*
* Then apply an affine transform which transforms [x_{k-1}, x_k] into
* [y_{k-1}, y_k]. The slope of this application is s_k.
* We impose a constraint on the  slope for this affine transform
*
* @f$ m_k=\cases{max(s_k, smin) & if $s_k <1$ \cr min(s_k, smax) & if $s_k >1$\cr} \f$
*
* @param data_out the array with the transformed initital array
* @param data initial array
* @param dim size of the array
* @param N number of control points (number of intervals of the partition - 1)
* @param smax maximum slope allowed
* @param smin minimum slope allowed
* @param points file with the points of the partition to draw the affine
* transformation
*
*/

int
piecewise_transformation(float *data_out, float *data, size_t dim, int N,
                         float smax, float smin, FILE * points) {

    float           x0, x1, y0, y1;
    float           Fu;
    float           min, max;
    int             k;
    float            *F;
    float           slope;

    if (NULL == (F = (float *) malloc(dim * sizeof(float)))) {
        fprintf(stderr, "allocation error\n");
        free(data);
        return EXIT_FAILURE;
    }


    minmax(data, dim, &min, &max);

    sort_data(data, dim, F);

    x0 = min;
    y0 = 0.;

    fprintf(points, "%i\n", N);
    for (k = 1; k <= N; k++) {
        Fu = ((float) (dim * k) / (float) (N + 1));
        y1 = (255. * (float) k) / (float) (N + 1);
        x1 = inverse_cumulative_function(Fu, F);
        if (x1 > x0) {
            slope = (y1 - y0) / (x1 - x0);

            if (slope > smax)
                y1 = smax * (x1 - x0) + y0;

            if (slope < smin)
                y1 = smin * (x1 - x0) + y0;


            fprintf(points, "%f %f %f %f\n", x0, x1, y0, y1);

            affine_transformation(data_out, data, dim, x0, x1, y0, y1);
            x0 = x1;
            y0 = y1;
        } else
            fprintf(points, "%f %f %f %f\n", x0, x0, y0, y0);
    }
    if (x0 < max) {
        y1 = 255.;
        x1 = max;
        slope = (y1 - y0) / (x1 - x0);

        if (slope > smax)
            y1 = smax * (x1 - x0) + y0;

        if (slope < smin)
            y1 = smin * (x1 - x0) + y0;



        fprintf(points, "%f %f %f %f\n", x0, x1, y0, y1);

        affine_transformation(data_out, data, dim, x0, x1, y0, y1);
        /*
         * if (y1 < 255.) printf(" with these slope parameters we don't reach
         * the \ maximal interval [0,255]\n"); if ( y1 > 255.) printf(" with
         * these slope parameters we obtain values \ outside [0,255]\n");
         */
    }

    else
        fprintf(points, "%f %f %f %f\n", x0, x0, y0, y0);

    free(F);
    return EXIT_SUCCESS;
}
